/**
 * @license https://github.com/Intermesh/goui/blob/main/LICENSE MIT License
 * @copyright Copyright 2023 Intermesh BV
 * @author Merijn Schering <mschering@intermesh.nl>
 */
import { Store, StoreComponent, StoreConfig, StoreEventMap, StoreRecord } from "../data/Store.js";
import { AbstractDataSource, BaseEntity, Changes, dataSourceEntityType, DefaultEntity, QueryParams, Filter } from "./AbstractDataSource.js";
import { ObservableListener } from "../component/index.js";
/**
 * Entity relation
 */
type Relation<EntityType extends BaseEntity> = Partial<Record<keyof EntityType, {
    /**
     * Data source to get relation from
     */
    dataSource: AbstractDataSource<EntityType>;
    /**
     * JSON pointer to relation key
     *
     * @see ObjectUtil.get()
     */
    path: string;
}>>;
type RecordBuilder<EntityType, StoreRecord> = (entity: EntityType) => Promise<StoreRecord>;
/**
 * DataSourceStore class
 *
 * Uses an {@see AbstractDataSource} to present data in the view.
 * @category Data
 */
export declare class DataSourceStore<DataSource extends AbstractDataSource = AbstractDataSource, RecordType extends StoreRecord = dataSourceEntityType<DataSource>> extends Store<RecordType> {
    dataSource: DataSource;
    queryParams: Omit<QueryParams, "sort">;
    hasMore: boolean;
    /**
     * En entity relations
     */
    relations?: Relation<DefaultEntity>;
    /**
     * Reload when the datasource changes
     */
    monitorChanges: boolean;
    /**
     * Builds record from entity
     * @param entity
     */
    buildRecord: RecordBuilder<dataSourceEntityType<DataSource>, RecordType>;
    constructor(dataSource: DataSource);
    private listening;
    /**
     * Reloads the store when the datasource changes
     *
     * @protected
     */
    protected onDSChange(DataSource: DataSource, changes: Changes): void;
    bindComponent(comp: StoreComponent<this, RecordType>): void;
    private listen;
    unbindComponent(comp: StoreComponent<this, RecordType>): void;
    protected internalLoad(append?: boolean): Promise<Awaited<RecordType>[]>;
    private fetchRelations;
    private keepPosition;
    reload(): Promise<RecordType[]>;
    load(append?: boolean): Promise<RecordType[]>;
    loadNext(append?: boolean): Promise<RecordType[]>;
    loadPrevious(): Promise<RecordType[]>;
    hasNext(): boolean;
    hasPrevious(): boolean;
    /**
     * Patch an existing filter with extra filter options
     *
     * @param ref
     * @param filter
     */
    patchFilter(ref: string, filter: Filter | undefined): this;
    private filters;
    /**
     * Set a filter for the {@see AbstractDataSource::query()} method
     *
     * @param ref A reference name so it can be replaced later by the component.
     * @param filter
     */
    setFilter(ref: string, filter: Filter | undefined): this;
    clearFilter(...names: string[]): void;
    getFilter(name: string): any;
}
export type DataSourceStoreConfig<DataSource extends AbstractDataSource, RecordType extends StoreRecord> = Omit<StoreConfig<RecordType>, "listeners"> & {
    queryParams?: QueryParams;
    dataSource: DataSource;
    buildRecord?: RecordBuilder<dataSourceEntityType<DataSource>, RecordType>;
    relations?: Relation<DefaultEntity>;
    listeners?: ObservableListener<StoreEventMap<DataSourceStore<DataSource, RecordType>, RecordType>>;
    filters?: Record<string, Record<string, any>>;
};
/**
 * Shorthand function to create a {@see DataSourceStore}
 *
 * @param config
 */
export declare const datasourcestore: <DataSource extends AbstractDataSource = AbstractDataSource<DefaultEntity>, RecordType extends StoreRecord = dataSourceEntityType<DataSource>>(config: DataSourceStoreConfig<DataSource, RecordType>) => DataSourceStore<DataSource, RecordType>;
export {};
//# sourceMappingURL=DataSourceStore.d.ts.map